#!/usr/bin/python

from __future__ import print_function

import email.parser
import io
import os
import re
import shutil
import subprocess
import sys
import tempfile
import unittest

from userdir_ldap.exceptions import UDFormatError
import userdir_ldap.gpg as userdir_gpg


data_dir = os.path.join(os.path.dirname(__file__), 'data')


class GPGTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.gnupghome = tempfile.mkdtemp()
        cls.old_environment = os.environ
        os.environ.clear()
        os.environ['GNUPGHOME'] = cls.gnupghome
        os.environ['SEQUOIA_KEY_STORE'] = cls.gnupghome
        os.environ['SEQUOIA_CERT_STORE'] = cls.gnupghome

        # Generate a new ephemeral key using sq
        pkey_filename = os.path.join(cls.gnupghome, 'testkey.asc')
        result = subprocess.run([
            'sq', 'key', 'generate',
            '--shared-key',
            '--userid', 'test key',
            '--cannot-authenticate',
            '--without-password',
            '--output', pkey_filename,
            '--rev-cert', os.path.join(cls.gnupghome, 'testkey-rev.asc'),
            '--expiration', '1d'], capture_output=True)
        subprocess.run(['sq', 'key', 'import', pkey_filename])

        cls.fingerprint = None
        for line in result.stderr.splitlines():
            if m := re.match(r'.*Fingerprint: ([A-F0-9]+)', str(line)):
                cls.fingerprint = m.group(1)
                break
        assert cls.fingerprint is not None

        with open(pkey_filename, 'rb') as pkey:
            cmd = subprocess.Popen(['gpg', '--import'],
                                   stdin=pkey,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        stdout, stderr = cmd.communicate()
        assert cmd.wait() in [0, 2], stderr

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.gnupghome)
        for k, v in cls.old_environment.items():
            os.environ[k] = v

    def test_checksig_inline(self):
        result = subprocess.run(['sq', 'sign',
                                 '--signer', self.fingerprint, '--cleartext'],
                                input=b'this is a test message',
                                capture_output=True)
        payload = result.stdout.decode('utf-8')

        userdir_gpg.ClearKeyrings()
        mail = email.parser.Parser().parsestr(payload)
        res, mime = userdir_gpg.gpg_check_email_sig(mail, paranoid=True)
        self.assertTrue(res.ok)
        self.assertEqual('this is a test message\n', res.text)
        self.assertFalse(mime)

        mail = email.parser.Parser().parsestr('some extra text\n' + payload)
        res, mime = userdir_gpg.gpg_check_email_sig(mail, paranoid=False)
        self.assertTrue(res.ok)
        self.assertEqual('this is a test message\n', res.text)
        self.assertFalse(mime)
        with self.assertRaises(UDFormatError):
            userdir_gpg.gpg_check_email_sig(mail, paranoid=True)

        mail = email.parser.Parser().parsestr(payload + 'some extra text\n')
        res, mime = userdir_gpg.gpg_check_email_sig(mail, paranoid=False)
        self.assertTrue(res.ok)
        self.assertEqual('this is a test message\n', res.text)
        self.assertFalse(mime)
        with self.assertRaises(UDFormatError):
            userdir_gpg.gpg_check_email_sig(mail, paranoid=True)

    def test_checksig_mime(self):
        payload = '''\
Date: Tue, 27 Oct 2020 20:07:56 +0100
From: foo@example.com
To: bar@example.com
Subject: test message
Message-Id: <something.unique@example.com>
MIME-Version: 1.0
Content-Type: multipart/signed; micalg=pgp-sha512;
        protocol="application/pgp-signature"; boundary="EeQfGwPcQSOJBaQU"
Content-Disposition: inline


--EeQfGwPcQSOJBaQU
Content-Type: text/plain; charset=us-ascii
Content-Disposition: inline

This is a test message.

--EeQfGwPcQSOJBaQU
Content-Type: application/pgp-signature; name="signature.asc"

#SIGNATURE#

--EeQfGwPcQSOJBaQU--
'''

        signed_text = b'''\
Content-Type: text/plain; charset=us-ascii\r
Content-Disposition: inline\r
\r
This is a test message.\r
'''

        result = subprocess.run(['sq', 'sign',
                                 '--signer', self.fingerprint,
                                 '--signature-file', '-'],
                                input=bytes(signed_text),
                                capture_output=True)
        signature = result.stdout.decode('utf-8')
        payload = payload.replace('#SIGNATURE#', signature)

        userdir_gpg.ClearKeyrings()
        mail = email.parser.Parser().parsestr(payload)
        res, mime = userdir_gpg.gpg_check_email_sig(mail, paranoid=True)
        self.assertTrue(res.ok)
        self.assertEqual(signed_text, res.text)
        self.assertTrue(mime)

    def test_unsigned(self):
        payload = '''\
This is not a signed message
'''
        mail = email.parser.Parser().parsestr(payload)
        with self.assertRaises(UDFormatError):
            userdir_gpg.gpg_check_email_sig(mail, paranoid=True)

    def test_badsig(self):
        result = subprocess.run(['sq', 'sign',
                                 '--signer', self.fingerprint, '--cleartext'],
                                input=b'this is a test message',
                                capture_output=True)
        payload = result.stdout.decode('utf-8')
        payload = payload.replace('test message', 'test\nmessage')

        userdir_gpg.ClearKeyrings()
        mail = email.parser.Parser().parsestr(payload)
        res, mime = userdir_gpg.gpg_check_email_sig(mail, paranoid=True)
        self.assertFalse(res.ok)
        self.assertEqual(res.why, "Verification of signature failed")
        self.assertFalse(mime)

    def test_encrypt(self):
        encrypted = userdir_gpg.GPGEncrypt(
            u"This is an encrypted message.\n",
            self.fingerprint)
        self.assertTrue(encrypted.startswith('-----BEGIN PGP MESSAGE-----\n'), encrypted)
        with self.assertRaises(Exception) as cm:
            userdir_gpg.GPGEncrypt(u"This is an encrypted message.\n", "0xDEADBEEF")
        self.assertEqual(cm.exception.args, ("No key found matching 0xDEADBEEF", ))

    def test_sigcheck(self):
        payload = '''\
Date: Tue, 27 Oct 2020 20:07:56 +0100
From: foo@example.com
To: bar@example.com
Subject: test message
Message-Id: <something.unique@example.com>
MIME-Version: 1.0
Content-Type: multipart/signed; micalg=pgp-sha512;
        protocol="application/pgp-signature"; boundary="EeQfGwPcQSOJBaQU"
Content-Disposition: inline


--EeQfGwPcQSOJBaQU
Content-Type: text/plain; charset=us-ascii
Content-Disposition: inline

This is a test message.

--EeQfGwPcQSOJBaQU
Content-Type: application/pgp-signature; name="signature.asc"

#SIGNATURE#

--EeQfGwPcQSOJBaQU--
'''

        signed_text = b'''\
Content-Type: text/plain; charset=us-ascii\r
Content-Disposition: inline\r
\r
This is a test message.\r
'''

        result = subprocess.run(['sq', 'sign',
                                 '--signer', self.fingerprint,
                                 '--signature-file', '-'],
                                input=bytes(signed_text),
                                capture_output=True)
        signature = result.stdout.decode('utf-8')
        payload = payload.replace('#SIGNATURE#', signature)

        cmd = subprocess.Popen(
            [sys.executable, os.path.join(os.path.dirname(__file__), '..', 'sigcheck')],
            stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = cmd.communicate(input=payload.encode('utf-8'))
        self.assertEqual(cmd.wait(), 0)
        self.assertEqual(stdout, b'Inspecting message <something.unique@example.com>\nMessage <something.unique@example.com> passed\n')
        self.assertEqual(stderr, b'')


if __name__ == '__main__':
    unittest.main()

# vim: set et sw=4 sts=4:
